package main.java.SignalPlanRecommender;

import java.io.IOException;
import java.util.Collection;

import opennlp.common.synsem.Sign;

import main.java.adapter.AimsunCaseAdapter;
import main.java.adapter.CaseAdapter;
import main.java.evaluate.CaseBaseEvaluator;
import main.java.network.TCPServer;

import jcolibri.casebase.CachedLinearCaseBase;
import jcolibri.casebase.LinearCaseBase;
import jcolibri.cbraplications.StandardCBRApplication;
import jcolibri.cbrcore.Attribute;
import jcolibri.cbrcore.CBRCase;
import jcolibri.cbrcore.CBRCaseBase;
import jcolibri.cbrcore.CBRQuery;
import jcolibri.cbrcore.Connector;
import jcolibri.connector.DataBaseConnector;
import jcolibri.evaluation.Evaluator;
import jcolibri.exception.ExecutionException;
import jcolibri.method.retrieve.RetrievalResult;
import jcolibri.method.retrieve.NNretrieval.NNConfig;
import jcolibri.method.retrieve.NNretrieval.NNScoringMethod;
import jcolibri.method.retrieve.NNretrieval.similarity.global.Average;
import jcolibri.method.retrieve.NNretrieval.similarity.local.Equal;
import jcolibri.method.retrieve.NNretrieval.similarity.local.Table;
import jcolibri.method.retrieve.selection.SelectCases;


public class SignalPlanRecommender implements StandardCBRApplication {

	//Decides whether or not the signal plan is calculated. Should be false when evaluating and finding weights.
	static boolean SIMULATING = false;
	//Whether or not info should be printed
	static boolean VERBOSE = false;
	Connector connector;
	CBRCaseBase caseBase;
	TCPServer server;
	CaseAdapter caseAdapter;
	SignalPlan signalPlan;
	String adaptedSolution;
	public NNConfig simConfig;
	
	Attribute seasonAttribute;
	Attribute weatherSituationAttribute;
	Attribute timeOfDayAttribute;
	Attribute eventAttribute;
	Attribute dayOfWeekAttribute;
	Attribute temperatureAttribute;
	Attribute frictionAttribute;
	
	//The solution of the case that the evaluation process has temporarily removed. Used for evaluating the proposed solution
	ProblemSolution crossValidationCaseSolution;
	
	public ProblemSolution getCrossValidationCaseSolution() {
		return crossValidationCaseSolution;
	}

	public void setCrossValidationCaseSolution(
			ProblemSolution crossValidationCaseSolution) {
		this.crossValidationCaseSolution = crossValidationCaseSolution;
	}

	public SignalPlanRecommender() {
		this.caseAdapter = new AimsunCaseAdapter();
		simConfig = new NNConfig();
		
		simConfig.setDescriptionSimFunction(new Average());
		seasonAttribute = new Attribute("season", ProblemDescription.class);
		simConfig.addMapping(seasonAttribute, new Table("SeasonSim"));
		weatherSituationAttribute = new Attribute("weatherSituation", ProblemDescription.class);
		simConfig.addMapping(weatherSituationAttribute, new Table("WeatherSituationSim"));
		timeOfDayAttribute = new Attribute("timeOfDay", ProblemDescription.class);
		simConfig.addMapping(timeOfDayAttribute, new Table("TimeOfDaySim"));
		eventAttribute = new Attribute("event", ProblemDescription.class);
		simConfig.addMapping(eventAttribute, new Equal());
		dayOfWeekAttribute = new Attribute("dayOfWeek", ProblemDescription.class);
		simConfig.addMapping(dayOfWeekAttribute, new Table("DayOfWeekSim"));
		temperatureAttribute = new Attribute("temperature", ProblemDescription.class); 
		simConfig.addMapping(temperatureAttribute, new Table("TemperatureSim"));
		frictionAttribute = new Attribute("friction", ProblemDescription.class); 
		simConfig.addMapping(frictionAttribute, new Table("FrictionSim"));

	}
	
	@Override
	public synchronized void cycle(CBRQuery query) throws ExecutionException {
		if (VERBOSE)
			System.out.println("Cycle");
		//The simfunction calculates the overall similarity based on the similarity of each attribute. 
		//Using Average() just take the sum of similarities and divides by n
		simConfig.setDescriptionSimFunction(new Average());
	
		//Execute NN
		Collection<RetrievalResult> eval = NNScoringMethod.evaluateSimilarity(caseBase.getCases(), query, simConfig);
		
		if (VERBOSE) {
			System.out.println("Retrieved Cases: ");
			for(RetrievalResult nse : eval) {
				System.out.println("Evaluation " + nse.getEval() + " id " + nse.get_case().getID());
			}
		}
		Collection<RetrievalResult> selectedCases = SelectCases.selectTopKRR(eval, 1);
		CBRCase cbrCase = null;
//		System.out.println("Selected Cases: ");
		for(RetrievalResult nse : selectedCases) {
//			System.out.println(nse);
			cbrCase = nse.get_case();
		}

		ProblemSolution ps = (ProblemSolution) cbrCase.getSolution();
		
		if (SIMULATING != true)
			jcolibri.evaluation.Evaluator.getEvaluationReport().addDataToSeries("Similarity", CaseBaseEvaluator.evaluateSolutionSimilarity(ps,this.getCrossValidationCaseSolution()));
		
		if (SIMULATING == true) {
			System.out.println("Calculating signal plan");
			SignalPlanCalculator signalPlanCalculator = new SignalPlanCalculator(ps, query.getDescription());
			signalPlanCalculator.calculateSignalPlanUsingWebster();
			this.adaptedSolution = this.caseAdapter.adaptToExternalSystem(signalPlanCalculator.signalPlan);
			System.out.println("get solution " + this.adaptedSolution);
		}
	}
	

	@Override
	public void configure() throws ExecutionException {
		try {
			//Emulate data base server
			main.java.database.HSQLDBserver.init();
			
			//Create a data base connector
			connector = new DataBaseConnector();
			//Init the db connector with the config file
			connector.initFromXMLfile(jcolibri.util.FileIO.findFile("main/java/database/databaseconfig.xml"));

			//Create a Linear case-base for in-memory organization
			caseBase = new CachedLinearCaseBase();
			
			// Create the dialogs
			//....
			
		} catch (Exception e) {
			throw new ExecutionException(e);
		}

	}

	@Override
	public synchronized CBRCaseBase preCycle() throws ExecutionException {
		// Load cases from connector into the case base
		caseBase.init(connector);
		// Print the cases
		System.out.println("Following cases are loaded");
		Collection<CBRCase> cases = caseBase.getCases();
		for(CBRCase c : cases) {
			System.out.println(c);
		}
		
		return caseBase;
	}
	

	@Override
	public void postCycle() throws ExecutionException {
		// TODO Auto-generated method stub
		

	}
	
	public void setWeights(double[] weights) {
		this.simConfig.setWeight(weatherSituationAttribute, weights[0]);
		this.simConfig.setWeight(timeOfDayAttribute, weights[1]);
		this.simConfig.setWeight(eventAttribute, weights[2]);
		this.simConfig.setWeight(dayOfWeekAttribute, weights[3]);
		this.simConfig.setWeight(temperatureAttribute, weights[4]);
		this.simConfig.setWeight(frictionAttribute, weights[5]);
		this.simConfig.setWeight(seasonAttribute, weights[6]);
	}

	public static void main(String args[]) throws IOException {
		SignalPlanRecommender recommender = new SignalPlanRecommender();
		double[] weights = {0.03937007874015748, 0.984251968503937, 0.29133858267716534, 0.2125984251968504, 0.7322834645669292, 0.007874015748031496, 0.007874015748031496};
		SIMULATING = true;
		try {
			recommender.server = new TCPServer(5000);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		ProblemDescription problemDescription;
		try {
			//Configure the application
			recommender.configure();
			recommender.preCycle();
			recommender.setWeights(weights);
			
			boolean cont = true;
			while(cont) {
				problemDescription = recommender.caseAdapter.adaptFromExternalSystem(recommender.server.receiveData());
//				for (ProblemDescription pd : problemDescriptions) {
				CBRQuery query = new CBRQuery();
	//				ProblemDescription pd = new ProblemDescription();
	//				pd.setDateTimeWeekInformation("lol");
	//				pd.setTrafficSituation(new TrafficSituation(1));
	//				pd.setWeatherSituation(WeatherSituation.Sunny);					
				query.setDescription(problemDescription);
				recommender.cycle(query);
				
				recommender.server.sendData(recommender.adaptedSolution);
				
				try {
					String outcome = null;
					outcome = recommender.server.receiveData();
					System.out.println("Outcome: " + outcome);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				
				
				recommender.postCycle();

				cont = true;
			}
			
			
		} catch(Exception e) {
			org.apache.commons.logging.LogFactory.getLog(SignalPlanRecommender.class).error(e);
			System.err.println(e.getMessage());
		}
		
	}

}
